"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var tslib_1 = require("tslib");
var api_1 = require("./api");
var model_1 = require("./model");
var memory_fs_1 = require("./memory-fs");
var utils_1 = require("./utils");
var CacheLevel;
(function (CacheLevel) {
    CacheLevel[CacheLevel["FILE_FULL"] = 0] = "FILE_FULL";
    CacheLevel[CacheLevel["DIR_SHALLOW"] = 1] = "DIR_SHALLOW";
    CacheLevel[CacheLevel["DIR_DEEP"] = 2] = "DIR_DEEP";
    CacheLevel[CacheLevel["DIR_CONTENT"] = 3] = "DIR_CONTENT";
    CacheLevel[CacheLevel["STATS"] = 4] = "STATS";
    CacheLevel[CacheLevel["GONE"] = 5] = "GONE";
})(CacheLevel || (CacheLevel = {}));
function isCachedFile(cached) {
    return cached === CacheLevel.FILE_FULL || cached === CacheLevel.GONE;
}
function isCachedDir(cached) {
    return cached === CacheLevel.DIR_CONTENT || cached === CacheLevel.DIR_DEEP || cached === CacheLevel.DIR_SHALLOW || cached === CacheLevel.GONE;
}
function isCachedDirDeep(cached) {
    return cached === CacheLevel.DIR_CONTENT || cached === CacheLevel.DIR_DEEP || cached === CacheLevel.GONE;
}
function isCachedDirContent(cached) {
    return cached === CacheLevel.DIR_CONTENT || cached === CacheLevel.GONE;
}
function nodesToMap(tree, accumulator) {
    if (accumulator === void 0) { accumulator = {}; }
    if (tree) {
        tree.forEach(function (node) {
            if (model_1.isDir(node)) {
                nodesToMap(node.children, accumulator);
            }
            accumulator[node.fullPath] = node;
        });
    }
    return accumulator;
}
function transformAndReport(oldMap, newMap) {
    var diff = {
        toAdd: [],
        toDelete: [],
        toChange: []
    };
    Object.keys(oldMap).forEach(function (oldPath) {
        var oldNode = oldMap[oldPath];
        var newNode = newMap[oldPath];
        if (!newNode) {
            diff.toDelete.push(oldNode);
            diff.toChange = diff.toChange.filter(function (path) { return path !== oldPath; });
        }
        else if (model_1.isFile(oldNode)) {
            if (model_1.isFile(newNode) && newNode.content === undefined) {
                newNode.content = oldNode.content;
            }
            else if (oldNode.content) {
                diff.toChange.push(oldPath);
            }
        }
    });
    Object.keys(newMap).forEach(function (newPath) {
        var oldNode = oldMap[newPath];
        var newNode = newMap[newPath];
        if (!oldNode) {
            diff.toAdd.push(newNode);
        }
    });
    return diff;
}
var MemFsForCache = /** @class */ (function (_super) {
    tslib_1.__extends(MemFsForCache, _super);
    function MemFsForCache() {
        return _super !== null && _super.apply(this, arguments) || this;
    }
    MemFsForCache.prototype.emitDiffEvents = function (toDelete, toAdd, toChange) {
        var _this = this;
        toDelete.forEach(function (node) {
            var type = (model_1.isDir(node) ? 'directory' : 'file') + "Deleted";
            _this.events.emit(type, { type: type, fullPath: node.fullPath });
        });
        toAdd.forEach(function (node) {
            if (model_1.isDir(node)) {
                _this.events.emit('directoryCreated', { type: 'directoryCreated', fullPath: node.fullPath });
            }
            else if (model_1.isFile(node)) {
                if (node.content !== undefined) {
                    _this.events.emit('fileCreated', {
                        type: 'fileCreated',
                        fullPath: node.fullPath,
                        newContent: node.content
                    });
                }
            }
            else {
                throw new Error('unknown node type');
            }
        });
        toChange.forEach(function (fullPath) { return _this.loadTextFile(fullPath).then(function (newContent) {
            _this.events.emit('fileChanged', { type: 'fileChanged', fullPath: fullPath, newContent: newContent });
        }); });
    };
    MemFsForCache.prototype.replaceChildrenSync = function (fullPath, newChildren) {
        var cachedTree = this.getDir(fullPath);
        var newChildrenMap = nodesToMap(newChildren);
        var _a = transformAndReport(nodesToMap(cachedTree.children), newChildrenMap), toDelete = _a.toDelete, toAdd = _a.toAdd, toChange = _a.toChange;
        cachedTree.children = cachedTree.children.filter(function (c) { return !~toDelete.indexOf(c); });
        cachedTree.children.forEach(function (c) {
            if (model_1.isFile(c) && ~toChange.indexOf(c.fullPath)) {
                var newChild = newChildrenMap[c.fullPath];
                if (model_1.isFile(newChild) && newChild.content !== undefined) {
                    c.content = newChild.content;
                }
            }
        });
        cachedTree.children = cachedTree.children.concat(toAdd);
        this.emitDiffEvents(toDelete, toAdd, toChange);
    };
    MemFsForCache.prototype.replaceDirSync = function (fullPath, newDir) {
        var cachedTree = this.getDir(fullPath);
        var newChildrenMap = nodesToMap(newDir.children);
        var _a = transformAndReport(nodesToMap(cachedTree.children), newChildrenMap), toDelete = _a.toDelete, toAdd = _a.toAdd, toChange = _a.toChange;
        var pathArr = utils_1.getPathNodes(fullPath);
        if (pathArr.length === 0) {
            this.root.children = newDir.children;
        }
        else {
            var parent = model_1.Directory.getSubDir(this.root, pathArr.slice(0, pathArr.length - 1));
            if (!parent) {
                throw new Error('directory parent missing: ' + pathArr.slice(0, pathArr.length - 1).join('/'));
            }
            else {
                parent.children = parent.children.map(function (child) { return child.name === newDir.name ? newDir : child; });
            }
        }
        this.emitDiffEvents(toDelete, toAdd, toChange);
    };
    return MemFsForCache;
}(memory_fs_1.MemoryFileSystem));
var CacheFileSystem = /** @class */ (function () {
    function CacheFileSystem(fs, options) {
        if (options === void 0) { options = {}; }
        var _this = this;
        this.options = options;
        this.cache = new MemFsForCache();
        this.events = this.cache.events;
        this.pathsInCache = {};
        this.onFsError = function (_a) {
            var stack = _a.stack;
            _this.options.reSyncOnError ?
                _this.reSyncOnError() :
                _this.emit('unexpectedError', { stack: stack });
        };
        this.fs = fs;
        if (this.options.reSyncOnError === undefined) {
            this.options.reSyncOnError = true;
        }
        if (this.options.propagateSyncRead && !api_1.isFileSystemReadSync(fs)) {
            throw new Error('propagateSyncRead option set to true but file system is not FileSystemReadSync');
        }
        this.baseUrl = fs.baseUrl;
        this.fs.events.on('unexpectedError', this.onFsError);
        var fileChangeCreatedHandler = function (event) {
            try {
                _this.cacheTextFile(event.fullPath, event.newContent, event.correlation);
            }
            catch (e) {
                _this.onFsError(e);
            }
        };
        this.fs.events.on('fileCreated', fileChangeCreatedHandler);
        this.fs.events.on('fileChanged', fileChangeCreatedHandler);
        this.fs.events.on('fileDeleted', function (event) {
            try {
                _this.cacheFileGone(event.fullPath, event.correlation);
            }
            catch (e) {
                _this.onFsError(e);
            }
        });
        this.fs.events.on('directoryCreated', function (event) {
            try {
                _this.cacheDirectoryExists(event.fullPath, event.correlation);
            }
            catch (e) {
                _this.onFsError(e);
            }
        });
        this.fs.events.on('directoryDeleted', function (event) {
            try {
                _this.cacheDirectoryGone(true, event.fullPath, event.correlation);
            }
            catch (e) {
                _this.onFsError(e);
            }
        });
    }
    CacheFileSystem.prototype.isPropagateSyncRead = function () {
        return this.options.propagateSyncRead || false;
    };
    CacheFileSystem.prototype.cacheTextFile = function (fullPath, newContent, correlation) {
        this.pathsInCache[fullPath] = CacheLevel.FILE_FULL;
        this.cache.saveFileSync(fullPath, newContent, correlation);
    };
    CacheFileSystem.prototype.cacheDirTree = function (fullPath, newTree) {
        this.pathsInCache[fullPath] = CacheLevel.DIR_DEEP;
        this.cache.ensureDirectorySync(fullPath);
        this.cache.replaceDirSync(fullPath, newTree);
    };
    CacheFileSystem.prototype.cacheDirChildren = function (fullPath, newChildren) {
        newChildren.forEach(function (c) {
            if (model_1.isDir(c)) {
                c.children = [];
            }
        });
        this.pathsInCache[fullPath] = CacheLevel.DIR_SHALLOW;
        this.cache.ensureDirectorySync(fullPath);
        this.cache.replaceChildrenSync(fullPath, newChildren);
    };
    CacheFileSystem.prototype.cacheStat = function (fullPath, newStat) {
        this.pathsInCache[fullPath] = CacheLevel.STATS;
        if (newStat.type === 'file') {
            this.cache.saveFileSync(fullPath, '');
        }
        else if (newStat.type === 'dir') {
            this.cache.ensureDirectorySync(fullPath);
        }
    };
    CacheFileSystem.prototype.cacheDirectoryGone = function (recursive, fullPath, correlation) {
        var _this = this;
        if (recursive) {
            Object.keys(this.pathsInCache).forEach(function (p) {
                if (p.startsWith(fullPath)) {
                    _this.pathsInCache[p] = CacheLevel.GONE;
                }
            });
        }
        else {
            this.pathsInCache[fullPath] = CacheLevel.GONE;
        }
        this.cache.deleteDirectorySync(fullPath, recursive, correlation);
    };
    CacheFileSystem.prototype.cacheDirectoryExists = function (fullPath, correlation) {
        if (this.pathsInCache[fullPath] === undefined || this.pathsInCache[fullPath] === CacheLevel.GONE) {
            this.pathsInCache[fullPath] = CacheLevel.DIR_DEEP;
        }
        this.cache.ensureDirectorySync(fullPath, correlation);
    };
    CacheFileSystem.prototype.cacheFileGone = function (fullPath, correlation) {
        this.pathsInCache[fullPath] = CacheLevel.GONE;
        this.cache.deleteFileSync(fullPath, correlation);
    };
    CacheFileSystem.prototype.saveFile = function (fullPath, newContent, correlation) {
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            return tslib_1.__generator(this, function (_a) {
                switch (_a.label) {
                    case 0: return [4 /*yield*/, this.fs.saveFile(fullPath, newContent, correlation)];
                    case 1:
                        correlation = _a.sent();
                        this.cacheTextFile(fullPath, newContent, correlation);
                        return [2 /*return*/, correlation];
                }
            });
        });
    };
    CacheFileSystem.prototype.deleteFile = function (fullPath, correlation) {
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            return tslib_1.__generator(this, function (_a) {
                switch (_a.label) {
                    case 0: return [4 /*yield*/, this.fs.deleteFile(fullPath, correlation)];
                    case 1:
                        correlation = _a.sent();
                        this.cacheFileGone(fullPath, correlation);
                        return [2 /*return*/, correlation];
                }
            });
        });
    };
    CacheFileSystem.prototype.deleteDirectory = function (fullPath, recursive, correlation) {
        if (recursive === void 0) { recursive = false; }
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            return tslib_1.__generator(this, function (_a) {
                switch (_a.label) {
                    case 0: return [4 /*yield*/, this.fs.deleteDirectory(fullPath, recursive, correlation)];
                    case 1:
                        correlation = _a.sent();
                        this.cacheDirectoryGone(recursive, fullPath);
                        return [2 /*return*/, correlation];
                }
            });
        });
    };
    CacheFileSystem.prototype.ensureDirectory = function (fullPath, correlation) {
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            return tslib_1.__generator(this, function (_a) {
                switch (_a.label) {
                    case 0: return [4 /*yield*/, this.fs.ensureDirectory(fullPath, correlation)];
                    case 1:
                        correlation = _a.sent();
                        this.cacheDirectoryExists(fullPath, correlation);
                        return [2 /*return*/, correlation];
                }
            });
        });
    };
    CacheFileSystem.prototype.loadTextFile = function (fullPath) {
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            var _a, _b;
            return tslib_1.__generator(this, function (_c) {
                switch (_c.label) {
                    case 0:
                        if (!!isCachedFile(this.pathsInCache[fullPath])) return [3 /*break*/, 2];
                        _a = this.cacheTextFile;
                        _b = [fullPath];
                        return [4 /*yield*/, this.fs.loadTextFile(fullPath)];
                    case 1:
                        _a.apply(this, _b.concat([_c.sent()]));
                        _c.label = 2;
                    case 2: return [2 /*return*/, this.cache.loadTextFileSync(fullPath)];
                }
            });
        });
    };
    CacheFileSystem.prototype.loadTextFileSync = function (fullPath) {
        if (this.isPropagateSyncRead() && !isCachedFile(this.pathsInCache[fullPath])) {
            this.cacheTextFile(fullPath, this.fs.loadTextFileSync(fullPath));
        }
        return this.cache.loadTextFileSync(fullPath);
    };
    CacheFileSystem.prototype.loadDirectoryTree = function (fullPath) {
        if (fullPath === void 0) { fullPath = ''; }
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            var _a, _b;
            return tslib_1.__generator(this, function (_c) {
                switch (_c.label) {
                    case 0:
                        if (!!isCachedDirDeep(this.pathsInCache[fullPath])) return [3 /*break*/, 2];
                        _a = this.cacheDirTree;
                        _b = [fullPath];
                        return [4 /*yield*/, this.fs.loadDirectoryTree(fullPath)];
                    case 1:
                        _a.apply(this, _b.concat([_c.sent()]));
                        _c.label = 2;
                    case 2: return [2 /*return*/, this.cache.loadDirectoryTreeSync(fullPath)];
                }
            });
        });
    };
    CacheFileSystem.prototype.loadDirectoryTreeSync = function (fullPath) {
        if (fullPath === void 0) { fullPath = ''; }
        if (this.isPropagateSyncRead() && !isCachedDirDeep(this.pathsInCache[fullPath])) {
            this.cacheDirTree(fullPath, this.fs.loadDirectoryTreeSync(fullPath));
        }
        return this.cache.loadDirectoryTreeSync(fullPath);
    };
    CacheFileSystem.prototype.loadDirectoryContentSync = function (fullPath) {
        if (fullPath === void 0) { fullPath = ''; }
        if (this.isPropagateSyncRead() && !isCachedDirContent(this.pathsInCache[fullPath])) {
            var content = this.fs.loadDirectoryContentSync(fullPath);
            this.pathsInCache[fullPath] = CacheLevel.DIR_CONTENT;
            this.cache.ensureDirectorySync(fullPath);
            this.cache.replaceDirSync(fullPath, model_1.Directory.fromContent(content));
        }
        return this.cache.loadDirectoryContentSync(fullPath);
    };
    CacheFileSystem.prototype.loadDirectoryChildren = function (fullPath) {
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            var _a, _b;
            return tslib_1.__generator(this, function (_c) {
                switch (_c.label) {
                    case 0:
                        if (!!isCachedDir(this.pathsInCache[fullPath])) return [3 /*break*/, 2];
                        _a = this.cacheDirChildren;
                        _b = [fullPath];
                        return [4 /*yield*/, this.fs.loadDirectoryChildren(fullPath)];
                    case 1:
                        _a.apply(this, _b.concat([_c.sent()]));
                        _c.label = 2;
                    case 2: return [2 /*return*/, this.cache.loadDirectoryChildrenSync(fullPath)];
                }
            });
        });
    };
    CacheFileSystem.prototype.loadDirectoryChildrenSync = function (fullPath) {
        if (this.isPropagateSyncRead() && !isCachedDir(this.pathsInCache[fullPath])) {
            this.cacheDirChildren(fullPath, this.fs.loadDirectoryChildrenSync(fullPath));
        }
        return this.cache.loadDirectoryChildrenSync(fullPath);
    };
    CacheFileSystem.prototype.stat = function (fullPath) {
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            var _a, _b;
            return tslib_1.__generator(this, function (_c) {
                switch (_c.label) {
                    case 0:
                        if (!(this.pathsInCache[fullPath] === undefined)) return [3 /*break*/, 2];
                        _a = this.cacheStat;
                        _b = [fullPath];
                        return [4 /*yield*/, this.fs.stat(fullPath)];
                    case 1:
                        _a.apply(this, _b.concat([_c.sent()]));
                        _c.label = 2;
                    case 2: return [2 /*return*/, this.cache.stat(fullPath)];
                }
            });
        });
    };
    CacheFileSystem.prototype.statSync = function (fullPath) {
        if (this.isPropagateSyncRead() && this.pathsInCache[fullPath] === undefined) {
            this.cacheStat(fullPath, this.fs.statSync(fullPath));
        }
        return this.cache.statSync(fullPath);
    };
    CacheFileSystem.prototype.dispose = function () {
        if (api_1.isDisposable(this.fs))
            this.fs.dispose();
    };
    CacheFileSystem.prototype.reSyncOnError = function () {
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            var _this = this;
            var oldPathsInCache;
            return tslib_1.__generator(this, function (_a) {
                oldPathsInCache = this.pathsInCache;
                this.pathsInCache = {};
                Object.keys(oldPathsInCache).forEach(function (p) {
                    if (oldPathsInCache[p] === CacheLevel.FILE_FULL) {
                        _this.loadTextFile(p);
                        _this.loadDirectoryChildren(utils_1.splitPathToDirAndFile(p).parentPath);
                    }
                    else if (oldPathsInCache[p] === CacheLevel.DIR_SHALLOW) {
                        _this.loadDirectoryChildren(p);
                        _this.loadDirectoryChildren(utils_1.splitPathToDirAndFile(p).parentPath);
                    }
                    else if (oldPathsInCache[p] === CacheLevel.DIR_DEEP) {
                        _this.loadDirectoryTree(p);
                        _this.loadDirectoryChildren(utils_1.splitPathToDirAndFile(p).parentPath);
                    }
                });
                return [2 /*return*/];
            });
        });
    };
    CacheFileSystem.prototype.emit = function (type, data) {
        this.events.emit(type, tslib_1.__assign({}, data, { type: type }));
    };
    return CacheFileSystem;
}());
exports.CacheFileSystem = CacheFileSystem;
//# sourceMappingURL=cache-fs.js.map