"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var tslib_1 = require("tslib");
var path = require("path");
var chokidar_1 = require("chokidar");
var promise_utils_1 = require("./promise-utils");
var api_1 = require("./api");
var model_1 = require("./model");
var local_fs_crud_only_1 = require("./local-fs-crud-only");
var utils_1 = require("./utils");
var events_manager_1 = require("./events-manager");
var notNoise = new WeakSet();
var LocalFileSystem = /** @class */ (function () {
    function LocalFileSystem(baseUrl, options) {
        if (options === void 0) { options = {
            interval: 100,
            retries: 3,
            correlationWindow: 200,
            noiseReduceWindow: 150
        }; }
        var _this = this;
        this.baseUrl = baseUrl;
        this.options = options;
        this.eventsManager = new events_manager_1.EventsManager();
        this.events = this.eventsManager.events;
        this.tempEventsTimers = {};
        this.lastEvents = {};
        this.crud = new local_fs_crud_only_1.LocalFileSystemCrudOnly(baseUrl);
        // events noise reduction
        this.eventsManager.addEventHandler({
            types: api_1.fileSystemEventNames,
            apply: function (ev) {
                if (_this.isDuplicatedEvent(ev)) {
                    return undefined;
                }
                else {
                    // clear any existing timer on suspicious file state
                    if (_this.tempEventsTimers[ev.fullPath] !== undefined) {
                        clearTimeout(_this.tempEventsTimers[ev.fullPath]);
                        delete _this.tempEventsTimers[ev.fullPath];
                    }
                    if (ev.type === 'fileChanged' && ev.newContent === '') {
                        // this is a suspicious file state. hide it for a while
                        _this.tempEventsTimers[ev.fullPath] = setTimeout(function () { return tslib_1.__awaiter(_this, void 0, void 0, function () {
                            var currentContent;
                            return tslib_1.__generator(this, function (_a) {
                                switch (_a.label) {
                                    case 0:
                                        delete this.tempEventsTimers[ev.fullPath];
                                        if (!(this.lastEvents[ev.fullPath] === ev)) return [3 /*break*/, 2];
                                        return [4 /*yield*/, this.loadTextFile(ev.fullPath)];
                                    case 1:
                                        currentContent = _a.sent();
                                        if (currentContent === ev.newContent) {
                                            // this event is not noise, re - emit it
                                            notNoise.add(ev);
                                            this.eventsManager.emit(ev);
                                        }
                                        _a.label = 2;
                                    case 2: return [2 /*return*/];
                                }
                            });
                        }); }, _this.options.noiseReduceWindow);
                        return undefined;
                    }
                    return ev;
                }
            },
            filter: function (ev) { return !notNoise.has(ev); }
        });
    }
    LocalFileSystem.prototype.isDuplicatedEvent = function (ev) {
        if (ev.fullPath) {
            var prevEv = this.lastEvents[ev.fullPath];
            this.lastEvents[ev.fullPath] = ev;
            return prevEv && (prevEv === ev || (prevEv.type === ev.type && prevEv.fullPath === ev.fullPath && prevEv.newContent === ev.newContent));
        }
        else {
            return false;
        }
    };
    LocalFileSystem.prototype.init = function () {
        var _this = this;
        var watcher = this.watcher = chokidar_1.watch([this.baseUrl], {
            // usePolling:true,
            // useFsEvents:false,
            // interval:50,
            //  atomic: false, //todo 50?
            cwd: this.baseUrl
        });
        watcher.on('error', function (err) { return console.log('Error in LocalFileSystem watch', err); });
        return new Promise(function (resolve) {
            watcher.once('ready', function () {
                watcher.on('addDir', function (relPath) {
                    if (relPath) { // ignore event of root folder creation
                        _this.eventsManager.emit({
                            type: 'directoryCreated',
                            fullPath: relPath.split(path.sep).join(model_1.pathSeparator)
                        });
                    }
                });
                watcher.on('add', function (relPath) { return tslib_1.__awaiter(_this, void 0, void 0, function () {
                    var _this = this;
                    var fullPath, e_1;
                    return tslib_1.__generator(this, function (_a) {
                        switch (_a.label) {
                            case 0:
                                fullPath = relPath.split(path.sep).join(model_1.pathSeparator);
                                _a.label = 1;
                            case 1:
                                _a.trys.push([1, 3, , 4]);
                                return [4 /*yield*/, promise_utils_1.retryPromise(function () { return tslib_1.__awaiter(_this, void 0, void 0, function () {
                                        var _a, _b, _c;
                                        return tslib_1.__generator(this, function (_d) {
                                            switch (_d.label) {
                                                case 0:
                                                    _b = (_a = this.eventsManager).emit;
                                                    _c = {
                                                        type: 'fileCreated',
                                                        fullPath: fullPath
                                                    };
                                                    return [4 /*yield*/, this.loadTextFile(relPath)];
                                                case 1: return [2 /*return*/, _b.apply(_a, [(_c.newContent = _d.sent(),
                                                            _c)])];
                                            }
                                        });
                                    }); }, this.options)];
                            case 2:
                                _a.sent();
                                return [3 /*break*/, 4];
                            case 3:
                                e_1 = _a.sent();
                                this.eventsManager.emit({ type: 'unexpectedError', fullPath: fullPath, stack: e_1.stack });
                                return [3 /*break*/, 4];
                            case 4: return [2 /*return*/];
                        }
                    });
                }); });
                watcher.on('change', function (relPath) { return tslib_1.__awaiter(_this, void 0, void 0, function () {
                    var _this = this;
                    var fullPath, e_2;
                    return tslib_1.__generator(this, function (_a) {
                        switch (_a.label) {
                            case 0:
                                fullPath = relPath.split(path.sep).join(model_1.pathSeparator);
                                _a.label = 1;
                            case 1:
                                _a.trys.push([1, 3, , 4]);
                                return [4 /*yield*/, promise_utils_1.retryPromise(function () { return tslib_1.__awaiter(_this, void 0, void 0, function () {
                                        var _a, _b, _c;
                                        return tslib_1.__generator(this, function (_d) {
                                            switch (_d.label) {
                                                case 0:
                                                    _b = (_a = this.eventsManager).emit;
                                                    _c = {
                                                        type: 'fileChanged',
                                                        fullPath: fullPath
                                                    };
                                                    return [4 /*yield*/, this.loadTextFile(relPath)];
                                                case 1: return [2 /*return*/, _b.apply(_a, [(_c.newContent = _d.sent(),
                                                            _c)])];
                                            }
                                        });
                                    }); }, this.options)];
                            case 2:
                                _a.sent();
                                return [3 /*break*/, 4];
                            case 3:
                                e_2 = _a.sent();
                                this.eventsManager.emit({ type: 'unexpectedError', fullPath: fullPath, stack: e_2.stack });
                                return [3 /*break*/, 4];
                            case 4: return [2 /*return*/];
                        }
                    });
                }); });
                watcher.on('unlinkDir', function (relPath) {
                    return _this.eventsManager.emit({
                        type: 'directoryDeleted',
                        fullPath: relPath.split(path.sep).join(model_1.pathSeparator)
                    });
                });
                watcher.on('unlink', function (relPath) {
                    return _this.eventsManager.emit({
                        type: 'fileDeleted',
                        fullPath: relPath.split(path.sep).join(model_1.pathSeparator)
                    });
                });
                resolve(_this);
            });
        });
    };
    LocalFileSystem.prototype.dispose = function () {
        this.watcher && this.watcher.close();
    };
    LocalFileSystem.prototype.saveFile = function (fullPath, newContent, correlation) {
        if (correlation === void 0) { correlation = utils_1.makeCorrelationId(); }
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            return tslib_1.__generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        this.registerCorrelationForPathsInDir(fullPath, correlation);
                        this.registerCorrelator(['fileChanged', 'fileCreated'], correlation, function (e) { return e.fullPath === fullPath && (e.newContent === newContent); }, true);
                        return [4 /*yield*/, this.crud.saveFile(fullPath, newContent)];
                    case 1:
                        _a.sent();
                        return [2 /*return*/, correlation];
                }
            });
        });
    };
    LocalFileSystem.prototype.deleteFile = function (fullPath, correlation) {
        if (correlation === void 0) { correlation = utils_1.makeCorrelationId(); }
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            return tslib_1.__generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        this.registerCorrelator(['fileDeleted'], correlation, function (e) { return e.fullPath === fullPath; }, true);
                        return [4 /*yield*/, this.crud.deleteFile(fullPath)];
                    case 1:
                        _a.sent();
                        return [2 /*return*/, correlation];
                }
            });
        });
    };
    LocalFileSystem.prototype.deleteDirectory = function (fullPath, recursive, correlation) {
        if (correlation === void 0) { correlation = utils_1.makeCorrelationId(); }
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            var prefix_1;
            return tslib_1.__generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        this.registerCorrelator(['directoryDeleted'], correlation, function (e) { return e.fullPath === fullPath; }, true);
                        if (recursive) {
                            prefix_1 = fullPath + model_1.pathSeparator;
                            this.registerCorrelator(['directoryDeleted', 'fileDeleted'], correlation, function (e) { return e.fullPath.startsWith(prefix_1); }, false);
                        }
                        return [4 /*yield*/, this.crud.deleteDirectory(fullPath, recursive)];
                    case 1:
                        _a.sent();
                        return [2 /*return*/, correlation];
                }
            });
        });
    };
    LocalFileSystem.prototype.ensureDirectory = function (fullPath, correlation) {
        if (correlation === void 0) { correlation = utils_1.makeCorrelationId(); }
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            return tslib_1.__generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        this.registerCorrelationForPathsInDir(fullPath, correlation);
                        this.registerCorrelator(['directoryCreated'], correlation, function (e) { return fullPath === e.fullPath; }, true);
                        return [4 /*yield*/, this.crud.ensureDirectory(fullPath)];
                    case 1:
                        _a.sent();
                        return [2 /*return*/, correlation];
                }
            });
        });
    };
    LocalFileSystem.prototype.loadTextFile = function (fullPath) {
        return this.crud.loadTextFile(fullPath);
    };
    LocalFileSystem.prototype.loadDirectoryTree = function (fullPath) {
        return this.crud.loadDirectoryTree(fullPath);
    };
    LocalFileSystem.prototype.loadDirectoryChildren = function (fullPath) {
        return this.crud.loadDirectoryChildren(fullPath);
    };
    LocalFileSystem.prototype.stat = function (fullPath) {
        return this.crud.stat(fullPath);
    };
    LocalFileSystem.prototype.registerCorrelationForPathsInDir = function (fullPath, correlation) {
        var nextPathSeparator = 0;
        var _loop_1 = function () {
            var subPath = fullPath.substr(0, nextPathSeparator);
            this_1.registerCorrelator(['directoryCreated'], correlation, function (e) { return subPath === e.fullPath; }, true);
        };
        var this_1 = this;
        while (~(nextPathSeparator = fullPath.indexOf(model_1.pathSeparator, nextPathSeparator + 1))) {
            _loop_1();
        }
    };
    LocalFileSystem.prototype.registerCorrelator = function (types, correlation, filter, single) {
        var _this = this;
        var correlator = {
            types: types,
            filter: filter,
            apply: function (e) {
                e.correlation = correlation;
                if (single) {
                    _this.eventsManager.removeEventHandler(correlator);
                }
                return e;
            },
        };
        this.eventsManager.addEventHandler(correlator, this.options.correlationWindow);
    };
    LocalFileSystem.prototype.loadTextFileSync = function (fullPath) {
        return this.crud.loadTextFileSync(fullPath);
    };
    LocalFileSystem.prototype.loadDirectoryTreeSync = function (fullPath) {
        return this.crud.loadDirectoryTreeSync(fullPath);
    };
    LocalFileSystem.prototype.loadDirectoryContentSync = function (fullPath) {
        return this.crud.loadDirectoryContentSync(fullPath);
    };
    LocalFileSystem.prototype.loadDirectoryChildrenSync = function (fullPath) {
        return this.crud.loadDirectoryChildrenSync(fullPath);
    };
    LocalFileSystem.prototype.statSync = function (fullPath) {
        return this.crud.statSync(fullPath);
    };
    return LocalFileSystem;
}());
exports.LocalFileSystem = LocalFileSystem;
//# sourceMappingURL=local-fs.js.map