"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var tslib_1 = require("tslib");
var model_1 = require("./model");
var utils_1 = require("./utils");
var id = 0;
var MemoryFileSystem = /** @class */ (function () {
    function MemoryFileSystem(baseUrl, options) {
        if (baseUrl === void 0) { baseUrl = "memory-" + id++; }
        this.baseUrl = baseUrl;
        this.events = utils_1.makeEventsEmitter();
        this.baseUrl += '/';
        if (options) {
            if (options.model && options.content) {
                throw new Error("MemoryFileSystem can't accept both model and content options");
            }
            this.root = options.model || model_1.Directory.fromContent(options.content || {});
        }
        else {
            this.root = model_1.Directory.fromContent({});
        }
    }
    MemoryFileSystem.addContent = function (fs, content, path) {
        var model = fs.root;
        var pathArr = path && utils_1.getPathNodes(path);
        if (path && pathArr && pathArr.length) {
            fs.ensureDirectorySync(path);
            var mixin = model_1.Directory.fromContent(content, pathArr[pathArr.length - 1], utils_1.normalizePathNodes(pathArr.slice(0, pathArr.length - 1)));
            model_1.Directory.mix(model_1.Directory.getSubDir(model, pathArr), mixin);
        }
        else {
            model_1.Directory.mix(model, model_1.Directory.fromContent(content));
        }
    };
    MemoryFileSystem.prototype.emit = function (type, event) {
        this.events.emit(type, event);
    };
    MemoryFileSystem.prototype.saveFile = function (fullPath, newContent, correlation) {
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            return tslib_1.__generator(this, function (_a) {
                return [2 /*return*/, this.saveFileSync(fullPath, newContent, correlation)];
            });
        });
    };
    MemoryFileSystem.prototype.deleteFile = function (fullPath, correlation) {
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            return tslib_1.__generator(this, function (_a) {
                return [2 /*return*/, this.deleteFileSync(fullPath, correlation)];
            });
        });
    };
    MemoryFileSystem.prototype.deleteDirectory = function (fullPath, recursive, correlation) {
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            return tslib_1.__generator(this, function (_a) {
                return [2 /*return*/, this.deleteDirectorySync(fullPath, recursive, correlation)];
            });
        });
    };
    MemoryFileSystem.prototype.ensureDirectory = function (fullPath, correlation) {
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            return tslib_1.__generator(this, function (_a) {
                return [2 /*return*/, this.ensureDirectorySync(fullPath, correlation)];
            });
        });
    };
    MemoryFileSystem.prototype.loadTextFile = function (fullPath) {
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            return tslib_1.__generator(this, function (_a) {
                return [2 /*return*/, this.loadTextFileSync(fullPath)];
            });
        });
    };
    MemoryFileSystem.prototype.loadDirectoryTree = function (fullPath) {
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            return tslib_1.__generator(this, function (_a) {
                return [2 /*return*/, this.loadDirectoryTreeSync(fullPath)];
            });
        });
    };
    MemoryFileSystem.prototype.loadDirectoryChildren = function (fullPath) {
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            return tslib_1.__generator(this, function (_a) {
                return [2 /*return*/, this.loadDirectoryChildrenSync(fullPath)];
            });
        });
    };
    MemoryFileSystem.prototype.stat = function (fullPath) {
        return tslib_1.__awaiter(this, void 0, void 0, function () {
            return tslib_1.__generator(this, function (_a) {
                return [2 /*return*/, this.statSync(fullPath)];
            });
        });
    };
    MemoryFileSystem.prototype.saveFileSync = function (fullPath, newContent, correlation) {
        if (correlation === void 0) { correlation = utils_1.makeCorrelationId(); }
        var pathArr = utils_1.getPathNodes(fullPath);
        var fileName = pathArr.pop();
        if (!fileName) {
            throw new Error("root is not a legal file name");
        }
        this._ensureDirectorySync(pathArr.join(model_1.pathSeparator), correlation);
        var parent = model_1.Directory.getSubDir(this.root, pathArr);
        if (!parent) {
            // we get here if findNode couldn't resolve the parent and the node is not the root dir.
            // this should not happen after running ensureDirectory()
            throw new Error("unexpected error: not a legal file name? '" + fullPath + "'");
        }
        var existingChild = parent.children.find(function (_a) {
            var name = _a.name;
            return name === fileName;
        });
        if (model_1.isDir(existingChild)) {
            throw new Error("file save error for path '" + fullPath + "'");
        }
        if (model_1.isFile(existingChild)) {
            if (existingChild.content !== newContent) {
                existingChild.content = newContent;
                var type = 'fileChanged';
                this.emit(type, { type: type, fullPath: fullPath, newContent: newContent, correlation: correlation });
            }
        }
        else {
            var type = 'fileCreated';
            parent.children.push(new model_1.File(fileName, fullPath, newContent));
            this.emit(type, { type: type, fullPath: fullPath, newContent: newContent, correlation: correlation });
        }
        return correlation;
    };
    MemoryFileSystem.prototype.deleteFileSync = function (fullPath, correlation) {
        if (correlation === void 0) { correlation = utils_1.makeCorrelationId(); }
        var pathArr = utils_1.getPathNodes(fullPath);
        var parent = pathArr.length ? model_1.Directory.getSubDir(this.root, pathArr.slice(0, pathArr.length - 1)) : null;
        if (model_1.isDir(parent)) {
            var node_1 = parent.children.find(function (_a) {
                var name = _a.name;
                return name === pathArr[pathArr.length - 1];
            });
            if (model_1.isFile(node_1)) {
                parent.children = parent.children.filter(function (_a) {
                    var name = _a.name;
                    return name !== node_1.name;
                });
                this.emit('fileDeleted', { type: 'fileDeleted', fullPath: fullPath, correlation: correlation });
            }
            else if (model_1.isDir(node_1)) {
                throw new Error("Directory is not a file '" + fullPath + "'");
            }
        }
        return correlation;
    };
    MemoryFileSystem.prototype.deleteDirectorySync = function (fullPath, recursive, correlation) {
        if (correlation === void 0) { correlation = utils_1.makeCorrelationId(); }
        var pathArr = utils_1.getPathNodes(fullPath);
        if (pathArr.length === 0) {
            throw new Error("Can't delete root directory");
        }
        var parent = model_1.Directory.getSubDir(this.root, pathArr.slice(0, pathArr.length - 1));
        if (model_1.isDir(parent)) {
            var node_2 = parent.children.find(function (_a) {
                var name = _a.name;
                return name === pathArr[pathArr.length - 1];
            });
            if (model_1.isFile(node_2)) {
                throw new Error("File is not a directory '" + fullPath + "'");
            }
            else if (model_1.isDir(node_2)) {
                if (!recursive && node_2.children.length) {
                    throw new Error("Directory is not empty '" + fullPath + "'");
                }
                else {
                    parent.children = parent.children.filter(function (_a) {
                        var name = _a.name;
                        return name !== node_2.name;
                    });
                    this.recursiveEmitDeletion(node_2, correlation);
                }
            }
        }
        return correlation;
    };
    MemoryFileSystem.prototype.ensureDirectorySync = function (fullPath, correlation) {
        if (correlation === void 0) { correlation = utils_1.makeCorrelationId(); }
        return this._ensureDirectorySync(fullPath, correlation);
    };
    MemoryFileSystem.prototype.findNode = function (fullPath) {
        var pathArr = utils_1.getPathNodes(fullPath);
        var parent = pathArr.length ? model_1.Directory.getSubDir(this.root, pathArr.slice(0, pathArr.length - 1)) : null;
        if (model_1.isDir(parent)) {
            var node = parent.children.find(function (_a) {
                var name = _a.name;
                return name === pathArr[pathArr.length - 1];
            });
            if (node) {
                return node;
            }
        }
        throw new Error("Cannot find " + fullPath);
    };
    MemoryFileSystem.prototype.getDir = function (fullPath) {
        var dir = model_1.Directory.getSubDir(this.root, fullPath);
        if (!dir) {
            throw new Error("Unable to read folder in path: '" + fullPath + "'");
        }
        return dir;
    };
    MemoryFileSystem.prototype.loadTextFileSync = function (fullPath) {
        var node = this.findNode(fullPath);
        if (model_1.isFile(node)) {
            return node.content || '';
        }
        else {
            throw new Error("File is a directory " + fullPath);
        }
    };
    MemoryFileSystem.prototype.loadDirectoryContentSync = function (fullPath) {
        if (fullPath === void 0) { fullPath = ''; }
        return model_1.Directory.toContent(this.getDir(fullPath));
    };
    MemoryFileSystem.prototype.loadDirectoryTreeSync = function (fullPath) {
        if (fullPath === void 0) { fullPath = ''; }
        return model_1.Directory.cloneStructure(this.getDir(fullPath));
    };
    MemoryFileSystem.prototype.loadDirectoryChildrenSync = function (fullPath) {
        return this.getDir(fullPath).children.map(function (child) { return model_1.isDir(child) ? new model_1.ShallowDirectory(child.name, child.fullPath) : new model_1.File(child.name, child.fullPath); });
    };
    MemoryFileSystem.prototype.statSync = function (fullPath) {
        var node = this.findNode(fullPath);
        return model_1.isFile(node) ?
            { type: 'file' } :
            { type: 'dir' };
    };
    MemoryFileSystem.prototype._ensureDirectorySync = function (fullPath, correlation) {
        var _this = this;
        utils_1.getPathNodes(fullPath).reduce(function (current, nodeName) {
            var next = current.children.find(function (_a) {
                var name = _a.name;
                return name === nodeName;
            });
            if (model_1.isDir(next)) {
                return next;
            }
            if (model_1.isFile(next)) {
                throw new Error("File is not a directory " + next.fullPath);
            }
            var newDir = new model_1.Directory(nodeName, current.fullPath ? [current.fullPath, nodeName].join(model_1.pathSeparator) : nodeName);
            current.children.push(newDir);
            _this.emit('directoryCreated', {
                type: 'directoryCreated',
                fullPath: newDir.fullPath,
                correlation: correlation
            });
            return newDir;
        }, this.root);
        return correlation;
    };
    MemoryFileSystem.prototype.recursiveEmitDeletion = function (node, correlation) {
        var _this = this;
        this.emit('directoryDeleted', { type: 'directoryDeleted', fullPath: node.fullPath, correlation: correlation });
        node.children.forEach(function (child) {
            if (model_1.isDir(child))
                _this.recursiveEmitDeletion(child, correlation);
            _this.emit('fileDeleted', { type: 'fileDeleted', fullPath: child.fullPath, correlation: correlation });
        });
    };
    return MemoryFileSystem;
}());
exports.MemoryFileSystem = MemoryFileSystem;
//# sourceMappingURL=memory-fs.js.map