"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var tslib_1 = require("tslib");
var chai_1 = require("chai");
var universal_1 = require("../src/universal");
var events_matcher_1 = require("./events-matcher");
exports.dirName = 'foo';
exports.fileName = 'bar.txt';
exports.content = 'content';
var testCorrelation = 'test-correlation';
function assertFileSystemContract(fsProvider, options) {
    describe("filesystem contract", function () {
        var fs;
        var matcher;
        beforeEach(function () {
            matcher = new events_matcher_1.EventsMatcher(options);
            return fsProvider()
                .then(function (newFs) {
                fs = newFs;
                matcher.track.apply(matcher, [fs.events].concat(universal_1.fileSystemEventNames));
            });
        });
        it("initially empty", function () {
            return chai_1.expect(fs.loadDirectoryTree()).to.become({ type: 'dir', name: '', fullPath: '', children: [] });
        });
        it("loading a non-existing file - fails", function () {
            return chai_1.expect(fs.loadTextFile(exports.fileName)).to.be.rejectedWith(Error);
        });
        it("loading a directory as a file - fails", function () {
            return fs.ensureDirectory(exports.dirName)
                .then(function (correlation) {
                return matcher.expect([{ type: 'directoryCreated', fullPath: exports.dirName, correlation: correlation }]);
            })
                .then(function () { return chai_1.expect(fs.loadTextFile(exports.dirName)).to.be.rejectedWith(Error); })
                .then(function () { return matcher.expect([]); });
        });
        it("stat a directory", function () {
            return tslib_1.__awaiter(this, void 0, void 0, function () {
                var type;
                return tslib_1.__generator(this, function (_a) {
                    switch (_a.label) {
                        case 0: return [4 /*yield*/, fs.ensureDirectory(exports.dirName)];
                        case 1:
                            _a.sent();
                            return [4 /*yield*/, fs.stat(exports.dirName)];
                        case 2:
                            type = (_a.sent()).type;
                            chai_1.expect(type).to.equal('dir');
                            return [2 /*return*/];
                    }
                });
            });
        });
        it("stat a file", function () {
            return tslib_1.__awaiter(this, void 0, void 0, function () {
                var type;
                return tslib_1.__generator(this, function (_a) {
                    switch (_a.label) {
                        case 0: return [4 /*yield*/, fs.saveFile(exports.fileName, exports.content)];
                        case 1:
                            _a.sent();
                            return [4 /*yield*/, fs.stat(exports.fileName)];
                        case 2:
                            type = (_a.sent()).type;
                            chai_1.expect(type).to.equal('file');
                            return [2 /*return*/];
                    }
                });
            });
        });
        it("stat a non-existing file/folder - fails", function () {
            return chai_1.expect(fs.stat(exports.fileName)).to.be.rejectedWith(Error);
        });
        it("saving an illegal file name - fails", function () {
            return chai_1.expect(fs.saveFile('', exports.content)).to.be.rejectedWith(Error)
                .then(function () {
                return matcher.expect([]);
            });
        });
        it("ensuring existence of directory", function () {
            var expectedStructure = {
                type: 'dir', name: '', fullPath: '', children: [
                    { type: 'dir', name: exports.dirName, fullPath: exports.dirName, children: [] }
                ]
            };
            return fs.ensureDirectory(exports.dirName)
                .then(function (correlation) { return matcher.expect([{ type: 'directoryCreated', fullPath: exports.dirName, correlation: correlation }]); })
                .then(function () { return chai_1.expect(fs.loadDirectoryTree()).to.become(expectedStructure); })
                .then(function () { return fs.ensureDirectory(exports.dirName); }) //2nd time does nothing
                .then(function () { return chai_1.expect(fs.loadDirectoryTree()).to.become(expectedStructure); })
                .then(function () { return matcher.expect([]); });
        });
        it("ensuring existence of directory (supplying correlation)", function () {
            var _this = this;
            var expectedStructure = {
                type: 'dir', name: '', fullPath: '', children: [
                    { type: 'dir', name: exports.dirName, fullPath: exports.dirName, children: [] }
                ]
            };
            return fs.ensureDirectory(exports.dirName, testCorrelation)
                .then(function (correlation) { return tslib_1.__awaiter(_this, void 0, void 0, function () {
                return tslib_1.__generator(this, function (_a) {
                    switch (_a.label) {
                        case 0:
                            chai_1.expect(correlation).to.equal(testCorrelation);
                            return [4 /*yield*/, matcher.expect([{ type: 'directoryCreated', fullPath: exports.dirName, correlation: correlation }])];
                        case 1:
                            _a.sent();
                            return [2 /*return*/];
                    }
                });
            }); })
                .then(function () { return chai_1.expect(fs.loadDirectoryTree()).to.become(expectedStructure); })
                .then(function () { return fs.ensureDirectory(exports.dirName); }) //2nd time does nothing
                .then(function () { return chai_1.expect(fs.loadDirectoryTree()).to.become(expectedStructure); })
                .then(function () { return matcher.expect([]); });
        });
        it("saving a file over a directory - fails", function () {
            return fs.ensureDirectory(exports.dirName)
                .then(function (correlation) { return matcher.expect([{ type: 'directoryCreated', fullPath: exports.dirName, correlation: correlation }]); })
                .then(function () { return chai_1.expect(fs.saveFile(exports.dirName, exports.content)).to.be.rejectedWith(Error); })
                .then(function () { return chai_1.expect(fs.loadDirectoryTree()).to.become({
                type: 'dir', name: '', fullPath: '', children: [
                    { type: 'dir', name: exports.dirName, fullPath: exports.dirName, children: [] }
                ]
            }); })
                .then(function () { return matcher.expect([]); });
        });
        it("saving a file over a file in its path - fails", function () {
            var fileNameAsDir = exports.dirName;
            return fs.saveFile(fileNameAsDir, exports.content)
                .then(function (correlation) { return matcher.expect([{
                    type: 'fileCreated',
                    fullPath: fileNameAsDir,
                    newContent: exports.content,
                    correlation: correlation
                }]); })
                .then(function () { return chai_1.expect(fs.saveFile(fileNameAsDir + "/" + exports.fileName, '_${content}')).to.be.rejectedWith(Error); })
                .then(function () { return chai_1.expect(fs.loadDirectoryTree()).to.become({
                type: 'dir', name: '', fullPath: '', children: [
                    { type: 'file', name: fileNameAsDir, fullPath: fileNameAsDir }
                ]
            }); })
                .then(function () { return matcher.expect([]); });
        });
        it("saving a new file (and a new directory to hold it)", function () {
            return fs.saveFile(exports.dirName + "/" + exports.fileName, exports.content)
                .then(function (correlation) { return matcher.expect([{ type: 'directoryCreated', fullPath: exports.dirName, correlation: correlation }, {
                    type: 'fileCreated',
                    fullPath: exports.dirName + "/" + exports.fileName,
                    newContent: exports.content,
                    correlation: correlation
                }]); })
                .then(function () { return chai_1.expect(fs.loadDirectoryTree()).to.become({
                type: 'dir', name: '', fullPath: '', children: [
                    {
                        type: 'dir', name: exports.dirName, fullPath: exports.dirName, children: [
                            { type: 'file', name: exports.fileName, fullPath: exports.dirName + "/" + exports.fileName }
                        ]
                    }
                ]
            }); })
                .then(function () { return matcher.expect([]); });
        });
        it("saving a file with different content", function () {
            var newContent = "_" + exports.content;
            return fs.saveFile(exports.fileName, exports.content)
                .then(function (correlation) { return matcher.expect([{
                    type: 'fileCreated',
                    fullPath: exports.fileName,
                    newContent: exports.content,
                    correlation: correlation
                }]); })
                .then(function () { return chai_1.expect(fs.loadTextFile(exports.fileName)).to.become(exports.content); })
                .then(function () { return fs.saveFile(exports.fileName, newContent); })
                .then(function (correlation) { return matcher.expect([{
                    type: 'fileChanged',
                    fullPath: exports.fileName,
                    newContent: newContent,
                    correlation: correlation
                }]); })
                .then(function () { return chai_1.expect(fs.loadTextFile(exports.fileName)).to.become(newContent); })
                .then(function () { return matcher.expect([]); });
        });
        it("saving a file with different content (supplying correlation)", function () {
            var newContent = "_" + exports.content;
            return fs.saveFile(exports.fileName, exports.content)
                .then(function (correlation) { return matcher.expect([{
                    type: 'fileCreated',
                    fullPath: exports.fileName,
                    newContent: exports.content,
                    correlation: correlation
                }]); })
                .then(function () { return chai_1.expect(fs.loadTextFile(exports.fileName)).to.become(exports.content); })
                .then(function () { return fs.saveFile(exports.fileName, newContent, testCorrelation); })
                .then(function (correlation) {
                chai_1.expect(correlation).to.equal(testCorrelation);
                return matcher.expect([{
                        type: 'fileChanged',
                        fullPath: exports.fileName,
                        newContent: newContent,
                        correlation: correlation
                    }]);
            })
                .then(function () { return chai_1.expect(fs.loadTextFile(exports.fileName)).to.become(newContent); })
                .then(function () { return matcher.expect([]); });
        });
        it("saving a file with same content", function () {
            var expectedStructure = {
                type: 'dir', name: '', fullPath: '', children: [
                    { name: exports.fileName, fullPath: exports.fileName, type: 'file' }
                ]
            };
            return fs.saveFile(exports.fileName, exports.content)
                .then(function (correlation) { return matcher.expect([{
                    type: 'fileCreated',
                    fullPath: exports.fileName,
                    newContent: exports.content,
                    correlation: correlation
                }]); })
                .then(function () { return chai_1.expect(fs.loadTextFile(exports.fileName)).to.become(exports.content); })
                .then(function () { return fs.saveFile(exports.fileName, exports.content); }) // may or may not trigger an event
                .then(function () { return chai_1.expect(fs.loadDirectoryTree()).to.become(expectedStructure); })
                .then(function () { return chai_1.expect(fs.loadTextFile(exports.fileName)).to.become(exports.content); });
        });
        it("deleting root directory - fails", function () {
            return chai_1.expect(fs.deleteDirectory('')).to.be.rejectedWith(Error)
                .then(function () { return matcher.expect([]); });
        });
        it("deleting a directory", function () {
            var _this = this;
            return fs.ensureDirectory(exports.dirName + "/_" + exports.dirName)
                .then(function (correlation) { return tslib_1.__awaiter(_this, void 0, void 0, function () {
                return tslib_1.__generator(this, function (_a) {
                    switch (_a.label) {
                        case 0:
                            fs;
                            return [4 /*yield*/, matcher.expect([
                                    { type: 'directoryCreated', fullPath: exports.dirName, correlation: correlation },
                                    { type: 'directoryCreated', fullPath: exports.dirName + "/_" + exports.dirName, correlation: correlation }
                                ])];
                        case 1:
                            _a.sent();
                            return [2 /*return*/];
                    }
                });
            }); })
                .then(function () { return fs.deleteDirectory(exports.dirName + "/_" + exports.dirName); })
                .then(function (correlation) { return tslib_1.__awaiter(_this, void 0, void 0, function () {
                return tslib_1.__generator(this, function (_a) {
                    switch (_a.label) {
                        case 0: return [4 /*yield*/, matcher.expect([{ type: 'directoryDeleted', fullPath: exports.dirName + "/_" + exports.dirName, correlation: correlation }])];
                        case 1:
                            _a.sent();
                            return [2 /*return*/];
                    }
                });
            }); })
                .then(function () { return chai_1.expect(fs.loadDirectoryTree()).to.eventually.have.property('children').eql([
                { children: [], fullPath: exports.dirName, name: exports.dirName, type: 'dir' }
            ]); })
                .then(function () { return matcher.expect([]); });
        });
        it("deleting a directory (supplying correlation)", function () {
            return fs.ensureDirectory(exports.dirName + "/_" + exports.dirName)
                .then(function (correlation) { return matcher.expect([
                { type: 'directoryCreated', fullPath: exports.dirName, correlation: correlation },
                { type: 'directoryCreated', fullPath: exports.dirName + "/_" + exports.dirName, correlation: correlation }
            ]); })
                .then(function () { return fs.deleteDirectory(exports.dirName + "/_" + exports.dirName, false, testCorrelation); })
                .then(function (correlation) {
                chai_1.expect(correlation).to.equal(testCorrelation);
                return matcher.expect([{ type: 'directoryDeleted', fullPath: exports.dirName + "/_" + exports.dirName, correlation: correlation }]);
            })
                .then(function () { return chai_1.expect(fs.loadDirectoryTree()).to.eventually.have.property('children').eql([
                { children: [], fullPath: exports.dirName, name: exports.dirName, type: 'dir' }
            ]); })
                .then(function () { return matcher.expect([]); });
        });
        it("deleting non existing directory succeeds", function () {
            return fs.deleteDirectory(exports.dirName + "/_" + exports.dirName)
                .then(function () { return chai_1.expect(fs.loadDirectoryTree()).to.eventually.have.property('children').eql([]); })
                .then(function () { return matcher.expect([]); });
        });
        it("deleting directory which is actually a file - fails", function () {
            return fs.saveFile(exports.fileName, exports.content)
                .then(function (correlation) { return matcher.expect([{
                    type: 'fileCreated',
                    fullPath: exports.fileName,
                    newContent: exports.content,
                    correlation: correlation
                }]); })
                .then(function () { return chai_1.expect(fs.deleteDirectory(exports.fileName)).to.be.rejectedWith(Error); })
                .then(function () { return matcher.expect([]); });
        });
        it("deleting non-empty directory without recursive flag - fails", function () {
            return fs.saveFile(exports.dirName + "/_" + exports.dirName + "/" + exports.fileName, exports.content)
                .then(function (correlation) { return matcher.expect([
                { type: 'directoryCreated', fullPath: exports.dirName, correlation: correlation },
                { type: 'directoryCreated', fullPath: exports.dirName + "/_" + exports.dirName, correlation: correlation },
                {
                    type: 'fileCreated',
                    fullPath: exports.dirName + "/_" + exports.dirName + "/" + exports.fileName,
                    newContent: exports.content,
                    correlation: correlation
                }
            ]); })
                .then(function () { return chai_1.expect(fs.deleteDirectory(exports.dirName + "/_" + exports.dirName)).to.be.rejectedWith(Error); })
                .then(function () { return matcher.expect([]); });
        });
        it("deleting non-empty directory with recursive flag", function () {
            var filePath = exports.dirName + "/_" + exports.dirName + "/" + exports.fileName;
            return fs.saveFile(filePath, exports.content)
                .then(function (correlation) { return matcher.expect([
                { type: 'directoryCreated', fullPath: exports.dirName, correlation: correlation },
                { type: 'directoryCreated', fullPath: exports.dirName + "/_" + exports.dirName, correlation: correlation },
                { type: 'fileCreated', fullPath: filePath, newContent: exports.content, correlation: correlation }
            ]); })
                .then(function () { return fs.deleteDirectory(exports.dirName, true); })
                .then(function (correlation) { return matcher.expect([
                { type: 'directoryDeleted', fullPath: exports.dirName, correlation: correlation },
                { type: 'fileDeleted', fullPath: filePath, correlation: correlation }
            ]); })
                .then(function () { return chai_1.expect(fs.loadDirectoryTree()).to.eventually.have.property('children').eql([]); })
                .then(function () { return matcher.expect([]); });
        });
        it("deleting file which is actually a directory - fails", function () {
            var dirNameAsFileName = exports.fileName;
            return fs.ensureDirectory(dirNameAsFileName)
                .then(function (correlation) { return matcher.expect([{
                    type: 'directoryCreated',
                    fullPath: dirNameAsFileName,
                    correlation: correlation
                }]); })
                .then(function () { return chai_1.expect(fs.deleteFile(dirNameAsFileName)).to.be.rejectedWith(Error); })
                .then(function () { return matcher.expect([]); });
        });
        it("deleting only one file", function () {
            return fs.saveFile(exports.fileName, exports.content)
                .then(function (correlation) { return matcher.expect([{
                    type: 'fileCreated',
                    fullPath: exports.fileName,
                    newContent: exports.content,
                    correlation: correlation
                }]); })
                .then(function () { return fs.saveFile("_" + exports.fileName, "_" + exports.content); })
                .then(function (correlation) { return matcher.expect([{
                    type: 'fileCreated',
                    fullPath: "_" + exports.fileName,
                    newContent: "_" + exports.content,
                    correlation: correlation
                }]); })
                .then(function () { return fs.deleteFile("_" + exports.fileName); })
                .then(function (correlation) { return matcher.expect([{ type: 'fileDeleted', fullPath: "_" + exports.fileName, correlation: correlation }]); })
                .then(function () { return chai_1.expect(fs.loadDirectoryTree()).to.eventually.have.property('children').eql([{
                    fullPath: exports.fileName,
                    name: exports.fileName,
                    type: 'file'
                }]); })
                .then(function () { return matcher.expect([]); });
        });
        it("deleting non existing file succeeds", function () {
            return fs.deleteFile(exports.fileName)
                .then(function () { return chai_1.expect(fs.loadDirectoryTree()).to.eventually.have.property('children').eql([]); })
                .then(function () { return matcher.expect([]); });
        });
        it("deleting non existing file (deep path) succeeds", function () {
            return fs.deleteFile(exports.dirName + "/" + exports.fileName)
                .then(function () { return chai_1.expect(fs.loadDirectoryTree()).to.eventually.have.property('children').eql([]); })
                .then(function () { return matcher.expect([]); });
        });
        it("loadDirectoryTree", function () {
            var expected = {
                fullPath: "", name: '', type: 'dir', children: [
                    {
                        fullPath: "" + exports.dirName, name: exports.dirName, type: 'dir', children: [
                            {
                                fullPath: exports.dirName + "/_" + exports.dirName, name: "_" + exports.dirName, type: 'dir', children: [
                                    { fullPath: exports.dirName + "/_" + exports.dirName + "/" + exports.fileName, name: exports.fileName, type: 'file' }
                                ]
                            }
                        ]
                    }
                ]
            };
            return fs.saveFile(exports.dirName + "/_" + exports.dirName + "/" + exports.fileName, exports.content)
                .then(function () { return chai_1.expect(fs.loadDirectoryTree()).to.eventually.eql(expected); })
                .then(function () { return chai_1.expect(fs.loadDirectoryTree(exports.dirName), "loadDirectoryTree('" + exports.dirName + "')").to.eventually.eql(expected.children[0]); })
                .then(function () { return chai_1.expect(fs.loadDirectoryTree(exports.dirName + "/_" + exports.dirName), "loadDirectoryTree('" + exports.dirName + "/_" + exports.dirName + "')").to.eventually.eql(expected.children[0].children[0]); });
        });
        it("loadDirectoryTree on an illegal sub-path", function () {
            return chai_1.expect(fs.loadDirectoryTree(exports.fileName)).to.be.rejectedWith(Error);
        });
        it("loadDirectoryChildren", function () {
            return fs.saveFile(exports.dirName + "/_" + exports.dirName + "/" + exports.fileName, exports.content)
                .then(function () { return fs.saveFile("" + exports.fileName, exports.content); })
                .then(function () { return chai_1.expect(fs.loadDirectoryChildren('')).to.eventually.have.deep.members([
                { fullPath: "" + exports.dirName, name: exports.dirName, type: 'dir' },
                { fullPath: exports.fileName, name: exports.fileName, type: 'file' }
            ]); })
                .then(function () { return chai_1.expect(fs.loadDirectoryChildren(exports.dirName), "loadDirectoryChildren('" + exports.dirName + "')").to.eventually.have.deep.members([{ fullPath: exports.dirName + "/_" + exports.dirName, name: "_" + exports.dirName, type: 'dir' }]); })
                .then(function () { return chai_1.expect(fs.loadDirectoryChildren(exports.dirName + "/_" + exports.dirName), "loadDirectoryChildren('" + exports.dirName + "/_" + exports.dirName + "')").to.eventually.have.deep.members([{ fullPath: exports.dirName + "/_" + exports.dirName + "/" + exports.fileName, name: exports.fileName, type: 'file' }]); });
        });
        it("loadDirectoryChildren on an illegal sub-path", function () {
            return chai_1.expect(fs.loadDirectoryChildren(exports.fileName)).to.be.rejectedWith(Error);
        });
        describe("correlation as argument", function () {
            var _this = this;
            var correlation = 'correlate this!';
            it("saveFile()", function () { return tslib_1.__awaiter(_this, void 0, void 0, function () {
                var correlationResult, correlation2, correlationResult2;
                return tslib_1.__generator(this, function (_a) {
                    switch (_a.label) {
                        case 0: return [4 /*yield*/, fs.saveFile(exports.dirName + "/" + exports.fileName, exports.content, correlation)];
                        case 1:
                            correlationResult = _a.sent();
                            chai_1.expect(correlationResult).to.eql(correlation);
                            return [4 /*yield*/, matcher.expect([{
                                        type: 'directoryCreated',
                                        fullPath: exports.dirName,
                                        correlation: correlation
                                    }, {
                                        type: 'fileCreated',
                                        fullPath: exports.dirName + "/" + exports.fileName,
                                        newContent: exports.content,
                                        correlation: correlation
                                    }])];
                        case 2:
                            _a.sent();
                            correlation2 = 'correlate that!';
                            return [4 /*yield*/, fs.saveFile(exports.dirName + "/" + exports.fileName, '', correlation2)];
                        case 3:
                            correlationResult2 = _a.sent();
                            chai_1.expect(correlationResult2).to.eql(correlation2);
                            return [4 /*yield*/, matcher.expect([{
                                        type: 'fileChanged',
                                        fullPath: exports.dirName + "/" + exports.fileName,
                                        newContent: '',
                                        correlation: correlation2
                                    }])];
                        case 4:
                            _a.sent();
                            return [2 /*return*/];
                    }
                });
            }); });
            it("deleteFile()", function () { return tslib_1.__awaiter(_this, void 0, void 0, function () {
                var correlationResult;
                return tslib_1.__generator(this, function (_a) {
                    switch (_a.label) {
                        case 0: return [4 /*yield*/, fs.saveFile(exports.fileName, exports.content)];
                        case 1:
                            _a.sent();
                            return [4 /*yield*/, matcher.expect([{
                                        type: 'fileCreated',
                                        fullPath: exports.fileName
                                    }])];
                        case 2:
                            _a.sent();
                            return [4 /*yield*/, fs.deleteFile(exports.fileName, correlation)];
                        case 3:
                            correlationResult = _a.sent();
                            chai_1.expect(correlationResult).to.eql(correlation);
                            return [4 /*yield*/, matcher.expect([{
                                        type: 'fileDeleted',
                                        fullPath: exports.fileName,
                                        correlation: correlation
                                    }])];
                        case 4:
                            _a.sent();
                            return [2 /*return*/];
                    }
                });
            }); });
            it("deleteDirectory()", function () { return tslib_1.__awaiter(_this, void 0, void 0, function () {
                var correlationResult;
                return tslib_1.__generator(this, function (_a) {
                    switch (_a.label) {
                        case 0: return [4 /*yield*/, fs.ensureDirectory(exports.dirName + "/" + exports.dirName)];
                        case 1:
                            _a.sent();
                            return [4 /*yield*/, matcher.expect([{
                                        type: 'directoryCreated',
                                        fullPath: exports.dirName + "/" + exports.dirName
                                    }])];
                        case 2:
                            _a.sent();
                            return [4 /*yield*/, fs.deleteDirectory(exports.dirName, true, correlation)];
                        case 3:
                            correlationResult = _a.sent();
                            chai_1.expect(correlationResult).to.eql(correlation);
                            return [4 /*yield*/, matcher.expect([{
                                        type: 'directoryDeleted',
                                        fullPath: exports.dirName,
                                        correlation: correlation
                                    }, {
                                        type: 'directoryDeleted',
                                        fullPath: exports.dirName + "/" + exports.dirName,
                                        correlation: correlation
                                    }])];
                        case 4:
                            _a.sent();
                            return [2 /*return*/];
                    }
                });
            }); });
            it("ensureDirectory()", function () { return tslib_1.__awaiter(_this, void 0, void 0, function () {
                var correlationResult;
                return tslib_1.__generator(this, function (_a) {
                    switch (_a.label) {
                        case 0: return [4 /*yield*/, fs.ensureDirectory(exports.dirName + "/" + exports.dirName, correlation)];
                        case 1:
                            correlationResult = _a.sent();
                            chai_1.expect(correlationResult).to.eql(correlation);
                            return [4 /*yield*/, matcher.expect([{
                                        type: 'directoryCreated',
                                        fullPath: exports.dirName,
                                        correlation: correlation
                                    }, {
                                        type: 'directoryCreated',
                                        fullPath: exports.dirName + "/" + exports.dirName,
                                        correlation: correlation
                                    }])];
                        case 2:
                            _a.sent();
                            return [2 /*return*/];
                    }
                });
            }); });
        });
        describe("action-event correlation", function () {
            it("single event per action", function () {
                return tslib_1.__awaiter(this, void 0, void 0, function () {
                    var correlation;
                    return tslib_1.__generator(this, function (_a) {
                        switch (_a.label) {
                            case 0:
                                this.timeout(30 * 1000);
                                return [4 /*yield*/, fs.saveFile(exports.fileName, 'foo')];
                            case 1:
                                correlation = _a.sent();
                                return [4 /*yield*/, matcher.expect([{ type: 'fileCreated', fullPath: exports.fileName, correlation: correlation }])];
                            case 2:
                                _a.sent();
                                return [4 /*yield*/, fs.saveFile(exports.fileName, 'bar')];
                            case 3:
                                correlation = _a.sent();
                                return [4 /*yield*/, matcher.expect([{ type: 'fileChanged', fullPath: exports.fileName, correlation: correlation }])];
                            case 4:
                                _a.sent();
                                return [4 /*yield*/, fs.deleteFile(exports.fileName)];
                            case 5:
                                correlation = _a.sent();
                                return [4 /*yield*/, matcher.expect([{ type: 'fileDeleted', fullPath: exports.fileName, correlation: correlation }])];
                            case 6:
                                _a.sent();
                                return [4 /*yield*/, fs.ensureDirectory(exports.dirName)];
                            case 7:
                                correlation = _a.sent();
                                return [4 /*yield*/, matcher.expect([{ type: 'directoryCreated', fullPath: exports.dirName, correlation: correlation }])];
                            case 8:
                                _a.sent();
                                return [4 /*yield*/, fs.deleteDirectory(exports.dirName)];
                            case 9:
                                correlation = _a.sent();
                                return [4 /*yield*/, matcher.expect([{ type: 'directoryDeleted', fullPath: exports.dirName, correlation: correlation }])];
                            case 10:
                                _a.sent();
                                return [2 /*return*/];
                        }
                    });
                });
            });
            it("multiple events per action", function () {
                return tslib_1.__awaiter(this, void 0, void 0, function () {
                    var correlation;
                    return tslib_1.__generator(this, function (_a) {
                        switch (_a.label) {
                            case 0:
                                this.timeout(10 * 1000);
                                return [4 /*yield*/, fs.saveFile(exports.dirName + "/" + exports.fileName, exports.content)];
                            case 1:
                                correlation = _a.sent();
                                chai_1.expect(correlation).to.be.a('string');
                                return [4 /*yield*/, matcher.expect([
                                        { type: 'directoryCreated', fullPath: exports.dirName, correlation: correlation },
                                        {
                                            type: 'fileCreated',
                                            fullPath: exports.dirName + "/" + exports.fileName,
                                            newContent: exports.content,
                                            correlation: correlation
                                        }
                                    ])];
                            case 2:
                                _a.sent();
                                return [4 /*yield*/, fs.deleteDirectory(exports.dirName, true)];
                            case 3:
                                correlation = _a.sent();
                                return [4 /*yield*/, matcher.expect([
                                        { type: 'directoryDeleted', fullPath: exports.dirName, correlation: correlation },
                                        {
                                            type: 'fileDeleted',
                                            fullPath: exports.dirName + "/" + exports.fileName,
                                            correlation: correlation
                                        }
                                    ])];
                            case 4:
                                _a.sent();
                                return [2 /*return*/];
                        }
                    });
                });
            });
        });
    });
}
exports.assertFileSystemContract = assertFileSystemContract;
function assertFileSystemSyncContract(fsProvider, options) {
    var fs;
    var matcher;
    beforeEach(function () {
        matcher = new events_matcher_1.EventsMatcher(options);
        return fsProvider()
            .then(function (newFs) {
            fs = newFs;
            matcher.track.apply(matcher, [fs.events].concat(universal_1.fileSystemEventNames));
        });
    });
    describe("filesystem sync contract", function () {
        var fs;
        var matcher;
        beforeEach(function () {
            matcher = new events_matcher_1.EventsMatcher(options);
            return fsProvider()
                .then(function (newFs) {
                fs = newFs;
                matcher.track.apply(matcher, [fs.events].concat(universal_1.fileSystemEventNames));
            });
        });
        it("loading a non-existing file - fails", function () {
            return chai_1.expect(function () { return fs.loadTextFileSync(exports.fileName); }).to.throw(Error);
        });
        it("loading an existing file - works", function () {
            return tslib_1.__awaiter(this, void 0, void 0, function () {
                return tslib_1.__generator(this, function (_a) {
                    switch (_a.label) {
                        case 0: return [4 /*yield*/, fs.saveFile(exports.fileName, exports.content)];
                        case 1:
                            _a.sent();
                            return [2 /*return*/, chai_1.expect(fs.loadTextFileSync(exports.fileName)).to.equal(exports.content)];
                    }
                });
            });
        });
        it("loading a directory as a file - fails", function () {
            return fs.ensureDirectory(exports.dirName)
                .then(function (correlation) {
                return matcher.expect([{ type: 'directoryCreated', fullPath: exports.dirName, correlation: correlation }]);
            })
                .then(function () { return chai_1.expect(function () { return fs.loadTextFileSync(exports.dirName); }).to.throw(Error); })
                .then(function () { return matcher.expect([]); });
        });
        it("statSync an existing directory", function () {
            return tslib_1.__awaiter(this, void 0, void 0, function () {
                var type;
                return tslib_1.__generator(this, function (_a) {
                    switch (_a.label) {
                        case 0: return [4 /*yield*/, fs.ensureDirectory(exports.dirName)];
                        case 1:
                            _a.sent();
                            type = fs.statSync(exports.dirName).type;
                            chai_1.expect(type).to.equal('dir');
                            return [2 /*return*/];
                    }
                });
            });
        });
        it("statSync an existing file", function () {
            return tslib_1.__awaiter(this, void 0, void 0, function () {
                var type;
                return tslib_1.__generator(this, function (_a) {
                    switch (_a.label) {
                        case 0: return [4 /*yield*/, fs.saveFile(exports.fileName, exports.content)];
                        case 1:
                            _a.sent();
                            type = fs.statSync(exports.fileName).type;
                            chai_1.expect(type).to.equal('file');
                            return [2 /*return*/];
                    }
                });
            });
        });
        it("statSync a non-existing file - fails", function () {
            return chai_1.expect(function () { return fs.statSync(exports.fileName); }).to.throw(Error);
        });
        it("loadDirectoryTreeSync", function () {
            var expected = {
                fullPath: "", name: '', type: 'dir', children: [
                    {
                        fullPath: "" + exports.dirName, name: exports.dirName, type: 'dir', children: [
                            {
                                fullPath: exports.dirName + "/_" + exports.dirName, name: "_" + exports.dirName, type: 'dir', children: [
                                    { fullPath: exports.dirName + "/_" + exports.dirName + "/" + exports.fileName, name: exports.fileName, type: 'file' }
                                ]
                            }
                        ]
                    }
                ]
            };
            return fs.saveFile(exports.dirName + "/_" + exports.dirName + "/" + exports.fileName, exports.content)
                .then(function () {
                chai_1.expect(fs.loadDirectoryTreeSync()).to.eql(expected);
                chai_1.expect(fs.loadDirectoryTreeSync(exports.dirName), "loadDirectoryTreeSync('" + exports.dirName + "')").to.eql(expected.children[0]);
                chai_1.expect(fs.loadDirectoryTreeSync(exports.dirName + "/_" + exports.dirName), "loadDirectoryTreeSync('" + exports.dirName + "/_" + exports.dirName + "')").to.eql(expected.children[0].children[0]);
            });
        });
        it("loadDirectoryTreeSync on an illegal sub-path", function () {
            return chai_1.expect(function () { return fs.loadDirectoryTreeSync(exports.fileName); }).to.throw(Error);
        });
        it("loadDirectoryChildrenSync", function () {
            return fs.saveFile(exports.dirName + "/_" + exports.dirName + "/" + exports.fileName, exports.content)
                .then(function () { return fs.saveFile("" + exports.fileName, exports.content); })
                .then(function () {
                chai_1.expect(fs.loadDirectoryChildrenSync('')).to.have.deep.members([
                    { fullPath: "" + exports.dirName, name: exports.dirName, type: 'dir' },
                    { fullPath: exports.fileName, name: exports.fileName, type: 'file' }
                ]);
                chai_1.expect(fs.loadDirectoryChildrenSync(exports.dirName), "loadDirectoryChildrenSync('" + exports.dirName + "')").to.have.deep.members([{ fullPath: exports.dirName + "/_" + exports.dirName, name: "_" + exports.dirName, type: 'dir' }]);
                chai_1.expect(fs.loadDirectoryChildrenSync(exports.dirName + "/_" + exports.dirName), "loadDirectoryChildrenSync('" + exports.dirName + "/_" + exports.dirName + "')").to.have.deep.members([{ fullPath: exports.dirName + "/_" + exports.dirName + "/" + exports.fileName, name: exports.fileName, type: 'file' }]);
            });
        });
        it("loadDirectoryChildrenSync on an illegal sub-path", function () {
            return chai_1.expect(function () { return fs.loadDirectoryChildrenSync(exports.fileName); }).to.throw(Error);
        });
        it("loadDirectoryContentSync", function () {
            var expected = universal_1.Directory.toContent({
                fullPath: "", name: '', type: 'dir', children: [
                    {
                        fullPath: "" + exports.dirName, name: exports.dirName, type: 'dir', children: [
                            {
                                fullPath: exports.dirName + "/_" + exports.dirName, name: "_" + exports.dirName, type: 'dir', children: [
                                    { fullPath: exports.dirName + "/_" + exports.dirName + "/" + exports.fileName, name: exports.fileName, type: 'file', content: exports.content }
                                ]
                            }
                        ]
                    }
                ]
            });
            return fs.saveFile(exports.dirName + "/_" + exports.dirName + "/" + exports.fileName, exports.content)
                .then(function () {
                chai_1.expect(fs.loadDirectoryContentSync()).to.eql(expected);
                chai_1.expect(fs.loadDirectoryContentSync(exports.dirName), "loadDirectoryContentSync('" + exports.dirName + "')").to.eql(expected[exports.dirName]);
                chai_1.expect(fs.loadDirectoryContentSync(exports.dirName + "/_" + exports.dirName), "loadDirectoryContentSync('" + exports.dirName + "/_" + exports.dirName + "')").to.eql(expected[exports.dirName]['_' + exports.dirName]);
            });
        });
        it("loadDirectoryTreeSync on an illegal sub-path", function () {
            return chai_1.expect(function () { return fs.loadDirectoryTreeSync(exports.fileName); }).to.throw(Error);
        });
    });
}
exports.assertFileSystemSyncContract = assertFileSystemSyncContract;
//# sourceMappingURL=implementation-suite.js.map