"use strict";
var _this = this;
Object.defineProperty(exports, "__esModule", { value: true });
var tslib_1 = require("tslib");
var tmp_1 = require("tmp");
var fs_1 = require("fs");
var path_1 = require("path");
var chai_1 = require("chai");
var implementation_suite_1 = require("./implementation-suite");
var events_matcher_1 = require("./events-matcher");
var nodejs_1 = require("../src/nodejs");
var eventMatcherOptions = {
    retries: 20,
    interval: 25,
    timeout: 1500,
    noExtraEventsGrace: 150
};
var fileSystemOptions = {
    interval: 100,
    retries: 3,
    correlationWindow: eventMatcherOptions.noExtraEventsGrace * 3,
    noiseReduceWindow: eventMatcherOptions.noExtraEventsGrace * 2 // eventMatcherOptions.timeout * 2.5
};
function writeFileToDisk(path, content) {
    fs_1.writeFileSync(path, content);
}
describe("the local filesystem implementation", function () {
    var dirCleanup;
    var rootPath;
    var testPath;
    var counter = 0;
    var disposableFileSystem;
    before(function (done) {
        tmp_1.dir({ unsafeCleanup: true }, function (_err, path, cleanupCallback) {
            dirCleanup = cleanupCallback;
            rootPath = path;
            done();
        });
    });
    after(function () {
        try {
            dirCleanup();
        }
        catch (e) {
            console.log('cleanup error', e);
        }
    });
    afterEach(function () {
        // if beforeEach fails, disposableFileSystem can stay undefined
        disposableFileSystem && disposableFileSystem.dispose();
    });
    function getFS() {
        testPath = path_1.join(rootPath, 'fs_' + (counter++));
        fs_1.mkdirSync(testPath);
        disposableFileSystem = new nodejs_1.LocalFileSystem(testPath, fileSystemOptions);
        return disposableFileSystem.init();
    }
    implementation_suite_1.assertFileSystemContract(getFS, eventMatcherOptions);
    implementation_suite_1.assertFileSystemSyncContract(getFS, eventMatcherOptions);
    describe("Local fs tests", function () {
        var fs;
        var matcher;
        beforeEach(function () { return tslib_1.__awaiter(_this, void 0, void 0, function () {
            return tslib_1.__generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        matcher = new events_matcher_1.EventsMatcher(eventMatcherOptions);
                        return [4 /*yield*/, getFS()];
                    case 1:
                        fs = _a.sent();
                        matcher.track.apply(matcher, [fs.events].concat(nodejs_1.fileSystemEventNames));
                        return [2 /*return*/];
                }
            });
        }); });
        describe("external changes", function () {
            it("handles dir creation", function () {
                var path = path_1.join(testPath, implementation_suite_1.dirName);
                fs_1.mkdirSync(path);
                return chai_1.expect(fs.loadDirectoryTree())
                    .to.eventually.have.property('children').eql([
                    { children: [], fullPath: implementation_suite_1.dirName, name: implementation_suite_1.dirName, type: 'dir' }
                ]);
            });
            it("handles dir deletion", function () {
                var path = path_1.join(testPath, implementation_suite_1.dirName);
                fs_1.mkdirSync(path);
                fs_1.rmdirSync(path);
                return chai_1.expect(fs.loadDirectoryTree()).to.eventually.have.property('children').eql([]);
            });
            it("handles file creation", function () {
                var path = path_1.join(testPath, implementation_suite_1.fileName);
                writeFileToDisk(path, implementation_suite_1.content);
                return chai_1.expect(fs.loadTextFile(implementation_suite_1.fileName)).to.eventually.equals(implementation_suite_1.content);
            });
            it("handles file deletion", function () {
                var path = path_1.join(testPath, implementation_suite_1.fileName);
                writeFileToDisk(path, implementation_suite_1.content);
                fs_1.unlinkSync(path);
                return chai_1.expect(fs.loadTextFile(implementation_suite_1.fileName)).to.eventually.be.rejected;
            });
            it("handles file change", function () { return tslib_1.__awaiter(_this, void 0, void 0, function () {
                var path, newContent, _a;
                return tslib_1.__generator(this, function (_b) {
                    switch (_b.label) {
                        case 0:
                            path = path_1.join(testPath, implementation_suite_1.fileName);
                            newContent = "_" + implementation_suite_1.content;
                            writeFileToDisk(path, implementation_suite_1.content);
                            return [4 /*yield*/, matcher.expect([{ type: 'fileCreated', fullPath: implementation_suite_1.fileName, newContent: implementation_suite_1.content }])];
                        case 1:
                            _b.sent();
                            writeFileToDisk(path, newContent);
                            _a = chai_1.expect;
                            return [4 /*yield*/, fs.loadTextFile(implementation_suite_1.fileName)];
                        case 2:
                            _a.apply(void 0, [_b.sent()]).to.equal(newContent);
                            return [2 /*return*/];
                    }
                });
            }); });
        });
        describe("events with 'newContent'", function () {
            it("emits 'unexpectedError' if 'loadTextFile' rejected in watcher 'add' callback", function () {
                fs.loadTextFile = function () { return Promise.reject('go away!'); };
                var path = path_1.join(testPath, implementation_suite_1.fileName);
                writeFileToDisk(path, implementation_suite_1.content);
                return matcher.expect([{ type: 'unexpectedError' }]);
            });
            it("emits 'unexpectedError' if 'loadTextFile' rejected in watcher 'change' callback", function () { return tslib_1.__awaiter(_this, void 0, void 0, function () {
                return tslib_1.__generator(this, function (_a) {
                    switch (_a.label) {
                        case 0: return [4 /*yield*/, fs.saveFile(implementation_suite_1.fileName, implementation_suite_1.content)];
                        case 1:
                            _a.sent();
                            return [4 /*yield*/, matcher.expect([{ type: 'fileCreated', fullPath: implementation_suite_1.fileName, newContent: implementation_suite_1.content }])];
                        case 2:
                            _a.sent();
                            fs.loadTextFile = function () { return Promise.reject('go away!'); };
                            return [4 /*yield*/, fs.saveFile(implementation_suite_1.fileName, "_" + implementation_suite_1.content)];
                        case 3:
                            _a.sent();
                            return [4 /*yield*/, matcher.expect([{ type: 'unexpectedError' }])];
                        case 4:
                            _a.sent();
                            return [2 /*return*/];
                    }
                });
            }); });
            it("emits exactly one 'change' event if 'loadTextFile' returns same content on multiple change events (unit for stress scenario)", function () { return tslib_1.__awaiter(_this, void 0, void 0, function () {
                var _this = this;
                var newContent;
                return tslib_1.__generator(this, function (_a) {
                    switch (_a.label) {
                        case 0: return [4 /*yield*/, fs.saveFile(implementation_suite_1.fileName, implementation_suite_1.content)];
                        case 1:
                            _a.sent();
                            return [4 /*yield*/, matcher.expect([{ type: 'fileCreated', fullPath: implementation_suite_1.fileName, newContent: implementation_suite_1.content }])];
                        case 2:
                            _a.sent();
                            newContent = "newContent";
                            fs.loadTextFile = function () { return tslib_1.__awaiter(_this, void 0, void 0, function () { return tslib_1.__generator(this, function (_a) {
                                return [2 /*return*/, newContent];
                            }); }); };
                            return [4 /*yield*/, fs.saveFile(implementation_suite_1.fileName, '123')];
                        case 3:
                            _a.sent();
                            return [4 /*yield*/, matcher.expect([{ type: 'fileChanged', fullPath: implementation_suite_1.fileName, newContent: newContent }])];
                        case 4:
                            _a.sent();
                            return [4 /*yield*/, fs.saveFile(implementation_suite_1.fileName, '456')];
                        case 5:
                            _a.sent();
                            return [4 /*yield*/, matcher.expect([])];
                        case 6:
                            _a.sent();
                            return [4 /*yield*/, fs.saveFile(implementation_suite_1.fileName, '789')];
                        case 7:
                            _a.sent();
                            return [4 /*yield*/, matcher.expect([])];
                        case 8:
                            _a.sent();
                            return [2 /*return*/];
                    }
                });
            }); });
        });
        describe('events noise', function () {
            var _this = this;
            nodejs_1.fileSystemEventNames.forEach(function (type) {
                it("de-dupe events of type " + type, function () { return tslib_1.__awaiter(_this, void 0, void 0, function () {
                    var ev1, ev2;
                    return tslib_1.__generator(this, function (_a) {
                        switch (_a.label) {
                            case 0:
                                ev1 = { type: type, fullPath: 'foo' };
                                ev2 = { type: type, fullPath: 'foo' };
                                fs.eventsManager.emit(ev1);
                                return [4 /*yield*/, matcher.expect([ev1])];
                            case 1:
                                _a.sent();
                                fs.eventsManager.emit(ev2);
                                return [4 /*yield*/, matcher.expect([])];
                            case 2:
                                _a.sent();
                                return [2 /*return*/];
                        }
                    });
                }); });
            });
            it('should dispatch events for empty files', function () { return tslib_1.__awaiter(_this, void 0, void 0, function () {
                var path;
                return tslib_1.__generator(this, function (_a) {
                    switch (_a.label) {
                        case 0:
                            path = path_1.join(testPath, implementation_suite_1.fileName);
                            writeFileToDisk(path, implementation_suite_1.content);
                            return [4 /*yield*/, matcher.expect([{ type: 'fileCreated', fullPath: implementation_suite_1.fileName, newContent: implementation_suite_1.content }])];
                        case 1:
                            _a.sent();
                            writeFileToDisk(path, '');
                            return [4 /*yield*/, matcher.expect([{ type: 'fileChanged', fullPath: implementation_suite_1.fileName, newContent: '' }])];
                        case 2:
                            _a.sent();
                            return [2 /*return*/];
                    }
                });
            }); });
            it('should not dispatch events for empty files if another change is detected within buffer time', function () { return tslib_1.__awaiter(_this, void 0, void 0, function () {
                var path;
                return tslib_1.__generator(this, function (_a) {
                    switch (_a.label) {
                        case 0:
                            path = path_1.join(testPath, implementation_suite_1.fileName);
                            writeFileToDisk(path, implementation_suite_1.content);
                            return [4 /*yield*/, matcher.expect([{ type: 'fileCreated', fullPath: implementation_suite_1.fileName, newContent: implementation_suite_1.content }])];
                        case 1:
                            _a.sent();
                            writeFileToDisk(path, '');
                            return [4 /*yield*/, matcher.expect([])];
                        case 2:
                            _a.sent();
                            writeFileToDisk(path, 'gaga');
                            return [4 /*yield*/, matcher.expect([{ type: 'fileChanged', fullPath: implementation_suite_1.fileName, newContent: 'gaga' }])];
                        case 3:
                            _a.sent();
                            return [2 /*return*/];
                    }
                });
            }); });
        });
    });
});
//# sourceMappingURL=local-fs.spec.node.js.map